//-----------------------------------------------------------------------------
//  DXL_typemaps.i  -  SWIG typemaps which change how SWIG parses the DXL.h
//                     prototypes for the Python DXL interface
//                   
//                     Included by DXL.i
//-----------------------------------------------------------------------------

%include typemaps.i

//------------------------------------------------------------------------------
//  Rename some of DX's functions so we can put our own Python versions
//    in their place.
//------------------------------------------------------------------------------

%rename DXLGetExecutionStatus           DXL_C_GetExecutionStatus;
%rename DXLSetSynchronization           DXL_C_SetSynchronization;
%rename DXLSetErrorHandler              DXL_C_SetErrorHandler;
%rename DXLSetMessageHandler            DXL_C_SetMessageHandler;
%rename DXLRemoveMessageHandler         DXL_C_RemoveMessageHandler;
%rename DXLSetValueHandler              DXL_C_SetValueHandler;
%rename DXLRemoveValueHandler           DXL_C_RemoveValueHandler;
%rename DXLSetBrokenConnectionCallback  DXL_C_SetBrokenConnectionCallback;

//----------------------------------------------------------
// Allow a None (NULL pointer) to specific pointer arguments
//----------------------------------------------------------

%typemap(python,in) const char *host
{
  if ( $source == Py_None )
    $target = NULL;
  else if ( PyString_Check( $source ) )
    $target = PyString_AsString( $source );
  else {
    PyErr_Format( PyExc_TypeError, 
               "argument 'host': expected None or string, %s found",
               $source == Py_None ? "None" : $source->ob_type->tp_name );
    return NULL;
  }
}

%typemap(python,in) const char *matchstr
{
  if ( $source == Py_None )
    $target = NULL;
  else if ( PyString_Check( $source ) )
    $target = PyString_AsString( $source );
  else {
    PyErr_Format( PyExc_TypeError, 
               "argument 'matchstr': expected None or string, %s found",
               $source == Py_None ? "None" : $source->ob_type->tp_name );
    return NULL;
  }
}

//----------------------------------------------------------------
// Allow a Python callable object in places expecting a C callback
//   Callbacks are a bit sticky to wrap because we have to 
//   maintain the C-back-to-Python mapping in the wrappers
//----------------------------------------------------------------

//  First, allow a Python callable object to be passed
//    FIXME:  May not need 
%typemap(python,in) DXLMessageHandler h, PyObject *pyfunc {
  if ( !PyCallable_Check( $source ) ) {
    PyErr_Format( PyExc_TypeError, 
               "argument 'h': expected callable object, %s found",
               $source == Py_None ? "None" : $source->ob_type->tp_name );
    return NULL;
  }
  $target = $source;
}

//-------------------------------------------------------
//  Certain input args are used to return output values
//-------------------------------------------------------

%apply int  *OUTPUT { int *state };

//-------------------------------------------------------
//  Turn C errors into Script Language exceptions
//-------------------------------------------------------

// NOTE:  PyDXLSyncMode is defined in DXL_additions.i.  Refer to it for details.

%{
#define PYDXL_SYNC_CHECK( conn ) \
          do {                                            \
            if ( PyDXLSyncMode )                          \
              if ( !WaitForDXIdleAndHandleError( conn ) ) \
                return NULL;                              \
          }                                               \
          while (0)                                       \
%}

%typemap(python,except) DXLConnection *DXLStartDX
{
  DXLConnection *ret = $function
  if ( !ret ) {
    PyErr_SetString( DXLExc_Error, "$name() failed" );
    return NULL;
  }
  PYDXL_SYNC_CHECK( ret );
}

// All functions returning DXLError take (DXLConnection *) as their first
//   arg, so using _arg0 let's us write one generic wrapper for all of them.
%typemap(python,except) DXLError
{
  DXLError ret;

  my_error_msg[0] = '\0';
  ret = $function

  if ( ret != OK ) {
    PyErr_Format( DXLExc_Error, "$name() failed%s%s", 
                  ( my_error_msg[0] ? ": " : "" ), my_error_msg );
    return NULL;
  }
  PYDXL_SYNC_CHECK( _arg0 );
}

// Special case - Waiting until DX is idle after closing the connection 
//   doesn't make any sense, so don't do it.
%typemap(python,except) DXLError DXLExitDX
{
  DXLError ret;

  my_error_msg[0] = '\0';
  ret = $function

  if ( ret != OK ) {
    PyErr_Format( DXLExc_Error, "$name() failed%s%s", 
                  ( my_error_msg[0] ? ": " : "" ), my_error_msg );
    return NULL;
  }
}

// FIXME:  exDXLLoadScript doesn't return meaningful DXLError values
%typemap(except) DXLError exDXLLoadScript
{
  $function
  PYDXL_SYNC_CHECK( _arg0 );
}

%typemap(except) void DXLCloseConnection, 
                 void DXLSetSynchronization,
                 void DXLSetSynchronization_NEW,
                 void DXLSetBrokenConnectionCallback
{
  $function
  PYDXL_SYNC_CHECK( _arg0 );
}

%typemap(python,except) DXLError DXLHandlePendingMessages
  /*  If a Python callback throws exceptions (e.g. SystemExit), we need to   */
  /*    record that fact and propagate it across the DXLink C layer.         */
{
  static int Reentrancy_check = 0;

  DXLError  ret;
  int       error = 0;

  /*  DXL, at first glance, "appears" to be callback reentrant.  But in case  */
  /*    not, flag recursive calls and investigate further if needed.          */
  if ( Reentrancy_check != 0 ) {
    fprintf( stderr, "$name() wrapper detected recursive invocation." );
    goto DONE;
  }
  Reentrancy_check = 1;

  PyErr_Clear();
  my_error_msg[0] = '\0';
  ret = $function

  if ( ret != OK ) {
    PyErr_Format( DXLExc_Error, "$name() failed%s%s", 
                  ( my_error_msg[0] ? ": " : "" ), my_error_msg );
    error = 1;
    goto DONE;
  }

  /*  If a Python callback threw an exception, restart its propagation  */
  /*    in the controlling Python.                                      */
  if ( PyErr_Occurred() ) {
    error = 1;
    goto DONE;
  }

  /*  Syncing here is somewhat useless, since DXLHandlePendingMessages     */
  /*    processes all the messages in the queue.  But we'll do it anyway.  */
  PYDXL_SYNC_CHECK( _arg0 );

 DONE:
  Reentrancy_check = 0;
  if ( error )
    return NULL;
}

%typemap(python,except) void WaitForDXIdle
{
  my_error        = 0;
  my_error_msg[0] = '\0';
  $function
  if ( my_error ) {
    PyErr_SetString( DXLExc_Error, my_error_msg );
    return NULL;
  }

  /*  If a Python callback threw an exception, restart its propagation  */
  /*    in the controlling Python.                                      */
  if ( PyErr_Occurred() )
    return NULL;
}

// A new version that returns status (see DXL_additions.i)
%typemap(python,except) int DXLGetExecutionStatus_NEW
{
  int status;

  my_error        = 0;
  my_error_msg[0] = '\0';
  status = $function
  if ( my_error ) {
    PyErr_SetString( DXLExc_Error, my_error_msg );
    return NULL;
  }
}
