#!/usr/bin/env python

import os, sys, types, time, DX
from   DXL import *

     # NOTE: Don't import Tkinter.  Let's not make DXServer GUI-specific.
     #       Importing it for the test() code below is OK, but not here.

class DXServer:
  """
  A thin wrapper around DXL to make it easier to use.

  See the test() routine in this module for a full example,
  but here's a brief sample usage:

     dx = DXServer()
     dx.SetDepth( depth )
     dx.SetHWRenderMode( hw_render_mode )
     dx.Open()
     dx.LoadVisualProgram( "mynet.net" )
     dx.ExecuteOnChange()
     dx.WaitForIdle()

     # On app termination
     dx.WaitForIdle()
     dx.Close()
  """

  def __init__( self ):
    ( self.conn, self.depth, self.hw_render_mode ) = ( None, 24, "none" )
    ( self.module_path, self.macro_path, self.mdf_file ) = [None]*3
    self.exec_on_change = 0
    self.command = DX.DXEXEC

  def SetDXCommand( self, command ):
    """  INPUT   : command        - Command which invokes IBM Data Explorer  """

    assert self.conn == None, "Connection is not closed."
    self.command = command

  def SetDepth( self, depth ):  
    """  INPUT   : depth          - depth of windows to display in  """

    assert self.conn == None, "Connection is not closed."
    self.depth = depth

  def SetHWRenderMode( self, hw_render_mode ):
    """  INPUT   : hw_render_mode - "none", "gl", "opengl"  """
                 
    assert hw_render_mode in [ "none", "gl","opengl" ]
    assert self.conn == None, "Connection is not closed."
    self.hw_render_mode = hw_render_mode

  def SetModulePath( self, module_path ):
    assert self.conn == None, "Connection is not closed."
    self.module_path = module_path

  def SetMacroPath( self, macro_path ):
    assert self.conn == None, "Connection is not closed."
    self.macro_path = macro_path

  def SetMDFFile( self, mdf_file ):
    assert self.conn == None, "Connection is not closed."
    self.mdf_file = mdf_file

  def Open( self ):
    assert self.conn == None, "Connection is not closed."

    self._ResetEnv()                # Get rid of all DX env var settings
    self._SetEnv()                  # And set only those we want

    # Build DX invocation command
    cmd = self.command + " -execonly"
    if self.hw_render_mode != "none": 
      cmd = cmd + " -hwrender %s" % self.hw_render_mode
    if self.depth == 8 : cmd = cmd + " -8bitcmap private"
    if self.mdf_file   : cmd = cmd + (" -mdf %s" % self.mdf_file)
    if self.module_path: cmd = cmd + (" -modules %s" % self.module_path)
    if self.macro_path : cmd = cmd + (" -macros %s" % self.macro_path)

    # Fire up DX
    #   FIXME:  Allow host specification someday
    self.conn = DXLStartDX( cmd, None )

    # By default, print errors and exit the program when they occur.
    #   Necessary because DX may be a zombie after an error occurs.
    #   (FIXME: One day, consider killing and restarting DX.  [...but
    #   then there's all that state to recreate.]
    self.SetErrorHandlerOnOff( 1 )

    # By default, print a message and exit if DX dies
    self.SetBrokenConnectionHandlerOnOff( 1 )

  def Close( self ):
    assert self.conn != None, "Connection is not open."
    DXLExitDX( self.conn )
    self.exec_on_change = 0

  def SetSynchronizationOnOff( self, enable ):
    """  Convenience wrapper for DXLSetSynchronization.  """
    
    # This calls a our version of the routine that works, unlike IBMs.
    #   It allows us to get error callbacks right after we commit the error
    DXLSetSynchronization( self.conn, enable )

  def SetMessageDebuggingOnOff( self, enable ):
    """  Convenience wrapper for DXLSetMessageDebugging.  """
    DXLSetMessageDebugging( self.conn, enable )

  def _DXBrokenConnectionHandlerCB( self, conn ):
    """  When enabled, called to handle DXLSetBrokenConnectionHandler()
         callback messages.
         This version just raises an exception.
         Enable error handler invocation via SetErrorHandlerOnOff()

         Override this in superclass if you want another broken connection
         handler.
    """

    raise DXError, "DX Connection Broken"

  def _DXErrorHandlerCB( self, conn, msg ):
    """  When enabled, called to handle DXLSetErrorHandler() callback messages.
         This version just prints the error to stdout and raises an exception.
         Enable error handler invocation via SetErrorHandlerOnOff()

         Error events == ERROR messages.  Both should be treated the same.
         When one is received, an exception should be thrown.  Then DX
         should be killed/restarted or just terminated because DX may be a
         zombie.  See comments in DXL_additions.i::WaitForDXIdle() for details.

         Override this in superclass if you want another error handler.
    """

    # NOTE:  Don't call any DXL routines in here.  If we're here because
    #   of a syntax error, DX is hosed and you could trigger another
    #   syntax error (even if your DXL calls are valid), which would
    #   cause this rtn to be called again.

    if self.error_handler_on:
      print "  <DXErr>       : %s" % msg

      # FIXME:
      # NOTE: Sadly, some messages sent to the DX error handler aren't
      #       crash-and-burn errors while some are.  This ugliness attempts
      #       to differentiate so we don't nuke the client needslessly.
      # EXAMPLES:
      #   "WARNING:  color map does not fully encompass data"
      if msg[:8] != "WARNING:":
        raise DXError, "Server Error: %s" % msg

  def _DXMessageHandlerCB( self, conn, msg, packet_type ):
    """  When enabled, default handler for received all DXLSetMessageHandler
         messages.
         This version just prints the messages to stdout...almost.

         If you plug in an ErrorHandler, and then you plug in an ERROR
         MessageHandler, it OVERRIDES the ErrorHandler (the ErrorHandler
         never gets called.  We fix this DXL annoyance by transferring all
         ERROR messages over to the error handler for processing.

         Error events == ERROR messages.  Both should be treated the same.
         When one is received, an exception should be thrown.  Then DX
         should be killed/restarted or just terminated because DX may be a
         zombie.  See comments in DXL_additions.i::WaitForDXIdle() for details.

         Enable message handler invocation via SetMessageHandlerOnOff()

         Override this in superclass if you want another message handler
    """

    msg_to_str_lookup = { PACK_INTERRUPT  : "INTERRUPT ",
                          PACK_MACRODEF   : "MACRODEF  ",
                          PACK_FOREGROUND : "FOREGROUND",
                          PACK_BACKGROUND : "BACKGROUND",
                          PACK_ERROR      : "ERROR     ",
                          PACK_MESSAGE    : "MESSAGE   ",
                          PACK_INFO       : "INFO      ",
                          PACK_LINQUIRY   : "LINQUIRY  ",
                          PACK_LRESPONSE  : "LRESPONSE ",
                          PACK_COMPLETE   : "COMPLETE  ",
                          PACK_LINK       : "LINK      " }

    # Hand off all ERROR events to the error handler.
    if packet_type == PACK_ERROR:
      print "Got Error via Message Handler"
      sys.stdout.flush()
      self._DXErrorHandlerCB( conn, msg )
    else:
      print "<DX Msg> %s:  %s" % ( msg_to_str_lookup[ packet_type ], msg )
      #global time1
      #try:              t = time.time() - time1
      #except NameError: (t,time1) = (0,time.time())
      #print "%8.2f: <DX Msg> %s:  %s" % \
      #       ( t, msg_to_str_lookup[ packet_type ], msg )

  def SetBrokenConnectionHandlerOnOff( self, enable ):
    if ( enable ): handler = self._DXBrokenConnectionHandlerCB
    else:          handler = None
    DXLSetBrokenConnectionCallback( self.conn, handler );  

  def SetErrorHandlerOnOff( self, enable ):

    # Always plug in a callback (in DXL, there is no "unplug" API)
    DXLSetErrorHandler( self.conn, self._DXErrorHandlerCB )
    self.error_handler_on = enable

  def SetMessageHandlerOnOff( self, enable, packet_types = None ):
    """  Used to enable/disable calling _DXErrorHandlerCB for error messages.

         NOTE: On new connections, a default PACK_ERROR handler is plugged in.
               But we can unplug it if we want, or plug something else in
               instead.

         ALSO NOTE:  Because of this, if you plug in an ErrorHandler, and
               then you plug in an ERROR MessageHandler, it OVERRIDES the
               ErrorHandler (the ErrorHandler never gets called.  We
               fix this DXL annoyance by transferring all ERROR
               messages over to the error handler for processing.

         INPUTS:  enable       - whether to enable or disable callbacks
                  packet_types - a list of the packet types to enable/disable
                                 printing messages for (if None, then
                                 enable/disable for all message types)
    """

    err_unplugged = 0
    
    def BuildCB( self, type ):
      def _( conn, msg, self=self, type=type ):
        self._DXMessageHandlerCB( conn, msg, type )
      return _

    if packet_types == None:
      packet_types = range( DXL_PACKETTYPE_LAST+1 )

    for i in packet_types:
      if enable:
        DXLSetMessageHandler   ( self.conn, i, None, BuildCB( self, i ) )
      else:
        DXLRemoveMessageHandler( self.conn, i, None )
        if i == PACK_ERROR:
          err_unplugged = 1

    # Make sure after it's all said and done that "some" handler is still
    #   plugged in to handle errors.  This is important!
    if err_unplugged:
      self.SetErrorHandlerOnOff( 1 )

  def _DXFileEventCB( self ):
    """ This callback should be invoked when activity is detected on the file
        descriptor associated with our DX connection.  The client
        determines when it is called.

        If you're familiar with XtAddInput() callbacks in Xt, this routine
        serves the same function.  Essentially when the inner select() of
        the application's main event loop says there is data available on
        the DX socket, call this routine to process it.

        E.g. For Tkinter, this routine can be hooked into the GUI's
        select() in its inner mainloop via createfilehandler().

        DX descriptor data is available when we're called, so process all
        pending DX messages and return.

        FIXME: Look at performance of this routine vs. frequency with which
               it is called and, if needed, consider implementing this in
               C like WaitForDXIdle

        INPUT  : file - a Python file for the DX connection
                 mask - the mask used when registering the file
                        (e.g. Tkinter.READABLE)

    """
    DXLHandlePendingMessages( self.conn );

  def _DXWorkProcCB( self ):
    """ This callback should be invoked periodically to process DX server
        messages queued but not yet delivered.

        Basically when DX receives a message from the server, it parses it,
        stuffs it into a structure, and puts it on an event queue
        (see _dxl_ReadFromSocket).

        In some circumstances, the events are all immediately processed
        (e.g. DXLHandlePendingMessages), invoking user callbacks as needed.
        However, in other scenarios, the events are queued but NOT
        processed (DXLSendPacket, DXLNextPacket, DXLGetPacket,
        DXLPeekPacket).

        Therefore we cannot simply use file descriptor activity on the DX
        socket descriptor as our queue to wake up and
        HandlePendingMessages.  This routine should be called periodically
        to process any 'queued but not processed' messages from the DX
        executive. 
       
        For more details, see the DXL implementation of InitializeXMainLoop
        and HandlePendingMessages.
    """
    DXLHandlePendingMessages( self.conn );

  def GetFileEventCB( self ):
    """ Returns a callable object taking no arguments to be invoked
        when descriptor activity is detected on the DX server socket.

        It cleans out the socket input buffer, parses and dispatches all
        available events, and then returns.

        RETURNS: callable object
    """

    return self._DXFileEventCB

  def GetWorkProcCB( self ):
    """ Returns a callable object taking no arguments to be invoked
        when periodically to dispatch 'queued but not processed' events in
        the DXL client layer.

        It cleans out the socket input buffer, parses and dispatches all
        available events, and then returns.

        RETURNS: callable object
    """

    return self._DXWorkProcCB

  def _ResetEnv( self ):
    """  Remove all DX variables from the environment.  DX behavior should
         be fully specified here and in the DX invocation script
         DXLStartDX invokes.
    """
    for var in os.environ.keys():
      if var[:2] == 'DX':
        del os.environ[ var ]

  def _SetEnv( self ):
    """  Set the DX env vars we want to be in effect for our DX executive.  """

    os.environ[ 'DXNO_BACKING_STORE' ] = "1"  # Make any difference??
    os.environ[ 'DXFLING'            ] = "1"  # Keep rotating after release

  #
  # Pass-through methods
  #
  def Sync( self ): DXLSync( self.conn )
  def WaitForIdle( self ): WaitForDXIdle( self.conn )
  def GetSocket  ( self ): return DXLGetSocket( self.conn )
  def Send  ( self, msg ): DXLSend( self.conn, msg )
  def BeginMacroDefinition( self, mhdr ):
    exDXLBeginMacroDefinition( self.conn, mhdr )
  def EndMacroDefinition( self ):
    exDXLEndMacroDefinition( self.conn )

  def GetExecutionStatus( self ): return DXLGetExecutionStatus( self.conn )
  def ExecuteOnce       ( self ):
    self.exec_on_change = 0
    DXLExecuteOnce    ( self.conn )
  def ExecuteOnChange   ( self ):
    self.exec_on_change = 1
    DXLExecuteOnChange( self.conn )
  def ExecuteOnceNamed    ( self, name ):
    self.exec_on_change = 0
    exDXLExecuteOnceNamed( self.conn, name )
  def ExecuteOnChangeNamed( self, name ):
    self.exec_on_change = 1
    exDXLExecuteOnChangeNamed( self.conn, name )
  def EndExecution      ( self ):
    self.exec_on_change = 0
    DXLEndExecution( self.conn )
  def EndExecuteOnChange( self ):
    self.exec_on_change = 0
    DXLEndExecuteOnChange( self.conn )

  # And a new state query
  def IsExecuteOnChangeActive( self ):
    return self.exec_on_change

  # APIs to set DXLInput and DXLInputNamed values
  #   (see the DXLInput note at bottom of file for why DXLSend is here too)
  def SetValue  ( self, var, value ):
    assert type(value) == types.StringType and value
    DXLSetValue  ( self.conn, var, value )
    DXLSend      ( self.conn, '%s = %s;' % (var,value) )
  def SetInteger( self, var, value ):
    DXLSetInteger( self.conn, var, value )
    DXLSend      ( self.conn, '%s = %s;' % (var,value) )
  def SetScalar ( self, var, value ):
    DXLSetScalar ( self.conn, var, value )
    DXLSend      ( self.conn, '%s = %s;' % (var,value) )
  def SetString ( self, var, value ):
    assert type(value) == types.StringType and value
    DXLSetString ( self.conn, var, value )
    DXLSend      ( self.conn, '%s = "%s";' % (var,value) )

  def LoadVisualProgram ( self, file ): DXLLoadVisualProgram ( self.conn, file )
  def LoadScript        ( self, file ): exDXLLoadScript( self.conn, file )

  def SequencerCtl( self, action ):
    DXLSequencerCtl( self.conn, action )

  #
  # Callback Registration Pass-through methods - 
  #   makes the first argument to the callback this DXServer instance rather
  #   than a DXLConnection.
  #
  def SetValueHandler( self, name, callback ):
    def BuildCB( self, callback ):
      def _( conn, name, value, callback=callback, self=self ):
        callback( self, name, value )
      return _

    DXLSetValueHandler( self.conn, name, BuildCB( self, callback ) )
    
  def RemoveValueHandler( self, name ):
    DXLRemoveValueHandler( self.conn, name )

  # NOT IMPLEMENTED FOR NOW (no need)
  def ExecuteOnceNamedWithArgs     ( self, name, *args ):
    raise NotImplementedError
  def ExecuteOnceNamedWithArgsV    ( self, name, args ):
    raise NotImplementedError
  def ExecuteOnChangeNamedWithArgs ( self, name, *args ):
    raise NotImplementedError
  def ExecuteOnChangeNamedWithArgsV( self, name, args ):
    raise NotImplementedError

  def Query( self, msg ):
    """ What would the client do with this?  DXL uses DXLQuery internally
        for DXLSync, DXLGetValue, and DXLGetExecutionStatus.
    """
    raise NotImplementedError

  # DXL APIs that shouldn't be exposed directly:
  #
  #  conn = DXLStartDX( cmd, host )
  #  conn = DXLConnectToRunningServer( port, host )
  #  bool = DXLIsMessagePending( conn )
  #  DXLExitDX( conn )
  #  DXLHandlePendingMessages( conn )
  #  DXLWaitForEvent( conn )
  #  DXLInitializeXMainLoop( xt_app_ctx, conn )
  #  DXLUninitializeXMainLoop( conn )




#------------------------------------------------------------------------------
#------------------------------------------------------------------------------


#
# TEST STUB
#
def test():

  from Tkinter  import *
  from DX       import DXEXEC

  # Defaults
  depth          = 24
  hw_render_mode = "opengl"

  # Create a server and open a connection
  print "\nFIRING UP DX:\n"
  dx = DXServer()
  dx.SetDXCommand( DXEXEC )
  dx.SetDepth( depth )
  dx.SetHWRenderMode( hw_render_mode )
  dx.Open()
  
  dx.SetMessageHandlerOnOff  ( 1 )    # Print all messages (default)
  dx.SetSynchronizationOnOff ( 1 )    # Chk msgs & wait idle for every DX call
  dx.SetErrorHandlerOnOff    ( 1 )    # Print errors and exit program (default)
  dx.SetMessageDebuggingOnOff( 1 )    # DX "stream-of-consciousness" prints

  # Fire up Tk
  root     = Tk()
  root.wm_withdraw()

  # Install a callback to handle DX messages in the Tk event loop
  def DXEventCB( file, mask, cb=dx.GetFileEventCB() ): cb()
  root.tk.createfilehandler( dx.GetSocket(), READABLE, DXEventCB )

  # Let DX get to idle (process queued messages, and empty message queue)
  dx.WaitForIdle()
  
  # Spin in the event loop for half a second
  print "\nSleeping for 1/2 second in the event loop.\n"
  root.after( 500, root.quit )
  root.mainloop()
  print "\nDone sleeping..\n"

  #---------------------------------------------------------------------------
  # SUCCESSFUL EXECUTION TEST
  #---------------------------------------------------------------------------
  
  print "\nSUCCESS TEST:  Try some useless work, like adding two numbers\n"

  # Turn off some of the verbosity now
  dx.SetMessageDebuggingOnOff( 0 )
  dx.SetMessageHandlerOnOff  ( 0 )

  # FIXME consider providing API on DXServer so conn is DXServer object
  def DXValueHandler( dx, name, value ):
    print "%s is %s" % ( name, value )

  print "Adding 1+2"
  dx.BeginMacroDefinition( 'macro main()' )
  dx.Send( "num1 = %d + %d;" % (1,2) )
  dx.Send( "DXLOutput( \"result\", num1 );" )
  dx.EndMacroDefinition()
  dx.SetValueHandler( "result", DXValueHandler )
  dx.ExecuteOnce()

  dx.WaitForIdle()
  dx.RemoveValueHandler( "result" )

  #---------------------------------------------------------------------------
  # FAILURE EXECUTION TEST
  #---------------------------------------------------------------------------
  
  print "\nFAILURE TEST:  Commit a syntax error via DXLSend()\n"

  # Now commit a DX error.
  #   Since synchronous mode is on, an exception will be thrown inside this
  #   DXLSend call.
  try:
    dx.BeginMacroDefinition( 'macro main()' )
    dx.Send( "num1 = 1 + 'abc';" )
    dx.Send( "DXLOutput( \"result\", num1 );" )
    dx.EndMacroDefinition()
    dx.ExecuteOnce()

    # Shut down DX
    dx.WaitForIdle()
    root.tk.deletefilehandler( dx.GetSocket() )
    dx.Close()
  except DXError:
    import traceback
    print "\nYep, we got that DXError exception we wanted.  " \
	  "Here's the traceback:\n"
    traceback.print_exc( file = sys.stdout )


  print "\nTESTING COMPLETE."

if __name__ == "__main__":
  test()

"""
-------------------------------------------------------------------------------
DX NOTES:

1) DXLINPUT

   DXLInputNamed and DXLInput are NOT functionally equivalent in all
   circumstances in a DX network (for commercial DX at least).

   A difference comes up when you're in ExecuteOnChange mode.  If you
   DXLSetInteger a DXLInput symbol for a .net-created DXLInput module, the
   network does NOT automatically re-execute in response to the change.  If
   you only use DXLSet*, you must force an execution in order for this
   change to propogate (DXLExecute*Once), or simply wait for the next
   execution to propagate it for it to take effect.  Not good.

   However, if you use a DXLInputNamed module instead, setting its symbol
   to a value via DXLSetInteger 'will' automatically re-execute the network
   like it's supposed to.

   THE HACK-AROUND: If you use DXLSend to set a DXLInput to a value, then
   it will actually work.  So for every DXLSet* call, we both call the
   appropriate DXLSet* call as well as invoke DXLSend to set the value of
   the script variable with the same name.  This covers both cases
   (i.e. where the module is a DXLInput, and where it's a DXLInputNamed).

   As to the robustness of this hack: Note that every DXLInput actually
   materializes as a script variable of the same name in the DX network
   anyway.  So when the module is a DXLInput, no naming conflicts are
   introduced by both DXLSending and DXLSeting the value.  In fact, DX is
   going to set this variable anyway so you might as well.  However, if the
   module is a 'DXLInputNamed', it doesn't materialize as a script
   variable.  So concievably there might be a conflict by setting a script
   variable of the same name using DXLSend.  The question is whether DX
   allows both a DXLInput and a DXLInputNamed which both answer to the same
   name.  Probably not since DXLSet can set both; either way, it's perverse
   to do this.  So by default our DXLSet* wrapper calls do both DXLSet and
   DXLSend for Set* wrapper calls.

   NOTES:

       1) For DXLInput, if you directly set the output script variable for the
          'virtual' DXLInput module (e.g. main_DXLInput_57_out_1 below) via
          DXLSetInteger, it still won't execute.  However, different than
          before, it 'won't' get picked up when you execute the net later
          either.

       2) DXLSetValue doesn't help you any over DXLSetInteger

   Interesting related note: When you do a DXLSet*, the DX message you get
   back across the wire explicitly says it's a SetDXLInputNamed event:

          <DX Msg> INTERRUPT :  begin /SetDXLInputNamed:0
          <DX Msg> INTERRUPT :  end   /SetDXLInputNamed:0

   Another interesting note (possibly related): In a .net file, notice that
   a DXLInputNamed module gets an explicit module definition:

          // 
          // node DXLInputNamed[3]: x = 4299, y = 424, inputs = 2, label = \
                  DXLInputNamed
          // input[1]: defaulting = 0, visible = 1, type = 32, value = \
                  "InteractorMode"
          // input[2]: defaulting = 0, visible = 1, type = 29, value = 0
          // page group: Main
          //
      main_DXLInputNamed_3_out_1 = 
          DXLInputNamed(
          main_DXLInputNamed_3_in_1,
          main_DXLInputNamed_3_in_2
          ) [instance: 3, cache: 1];
          // 
          // node Receiver[277]: x = 392, y = 98, inputs = 1, label = \
                  PickIsActive
          // page group: IRR_MB
          //

   whereas DXLInput does not.  It only gets a script variable placeholder to
   transfer to the output name of a 'virtual' module:

          // 
          // node DXLInput[57]: x = 4301, y = 394, inputs = 1, label = \
                  InteractorMode
          // input[1]: defaulting = 0, visible = 1, type = 29, value = 0
          // page group: Main
          //
          main_DXLInput_57_out_1 = InteractorMode;
          // 



-------------------------------------------------------------------------------
"""
